/*
DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
                Version 2, December 2004

Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>

Everyone is permitted to copy and distribute verbatim or modified
copies of this license document, and changing it is allowed as long
as the name is changed.

        DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

0. You just DO WHAT THE FUCK YOU WANT TO.

*/

/*
A very simple GUI display for the Hokuyo laser
*/

#ifndef __DISTANCETRANSFORM_H__
#define __DISTANCETRANSFORM_H__

#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <vector>
#include <iostream>

using namespace std;

// You can change this value
const int LINEARISATION_STEP = 30; // pixels, maximum length per linearise segment
const float MIN_DISTANCE_FROM_OBSTACLES = 10; // pixels, keep this much distance away from the walls


// Some constants for different cell states
const int UNVISITED = 0; // unvisited cell, used by GetNeighbours
const int VISITED = 1; // visited cell, used by GetNeighbours
const int ASSIGNED = 2; // has a value assigned by GetDistance
const int OBSTACLE = 4;
const int FREEPSACE = 5;

// Data structure to hold X,Y values
struct XY
{
    int x, y;

    XY(){}
    XY(int _x, int _y)
    {
        x = _x;
        y = _y;
    }
};

class DistanceTransform
{
public:
    DistanceTransform();
    ~DistanceTransform();

    void LoadMap(IplImage *in); // Load the map
    void CalcDistanceTransform(bool safePath = false); // Basic distance transform
    void FindPath(); // Find a path using the Distance transform
	void FindLinearisePath(); // Linearise path

    int w, h; // width and height of the maps
    char *map; // original binary map
    float *DTmap; // distance transformed map
    float *PFmap; // potential field map
    vector <XY> path; // the calculated path
    XY goal; // goal point
    XY start; // starting point
    float maxDTvalue; // maximum value from the Distance Transform, use for display purpose
    float maxPFvalue; // maximum value from the potential field

private:
    float GetDistance(float *Tmap, int x, int y); // smallest distance/cost to cell (x,y)
    void GetNeighbours(int x, int y, vector <XY> &neighbours); // get UNVISITED cells around (x,y)
    bool IsGoodPath(const XY &start, const XY &end); // Used for finding safe path
    void CalcPotentialField();
    char *cellState; // cell state map, same dimension as the map
};

#endif
