/*
DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
                Version 2, December 2004

Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>

Everyone is permitted to copy and distribute verbatim or modified
copies of this license document, and changing it is allowed as long
as the name is changed.

        DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

0. You just DO WHAT THE FUCK YOU WANT TO.

*/

/*
A demo application showing the distance transform for path planning
*/

#include <iostream>
#include <cstdio>
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <GL/glut.h>
#include "DistanceTransform.h"

const char MAP_FILE[] = "maze.bmp";

DistanceTransform DT;

// Normal = map + distance transform, DTMAP = colourised distance transform map
// User can choose by pressing '1' or '2'
enum DisplayType {NORMAL, DTMAP, PFMAP};
DisplayType DISPLAY_MODE = NORMAL;
bool normalPath; // displaying normal or linearise path

// GLUT functions
void reshape(int w, int h);
void draw();
void idle();
void keyboard(unsigned char key, int x, int y);
void mouse(int button, int state, int x, int y);
void renderBitmapString(float x, float y, float z, void *font, const char *string);
int SCREEN_W, SCREEN_H;


int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);

    // When we load images it will be upside down when display, so we'll flip it
    IplImage *map = cvLoadImage(MAP_FILE);

    if(!map) {
        cerr << "Error loading " << MAP_FILE << endl;
        return 1;
    }

    cvFlip(map, map); // upside down
    DT.LoadMap(map);
    cvReleaseImage(&map);

    glutInitWindowSize(DT.w, DT.h);

    glutCreateWindow("Distance transform");

	glutDisplayFunc(draw);
    glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);

	glutMainLoop();

    return 0;
}

void draw()
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Draw the map
    glPointSize(1.0);

    if(DISPLAY_MODE == NORMAL) { // Display the map and path
        glBegin(GL_POINTS);
        for(int y=0; y < DT.h; y++) {
            for(int x=0; x < DT.w; x++) {
                if(DT.map[y*DT.w+x] == OBSTACLE) {
                    glColor3f(0,0,0);
                    glVertex2i(x,y);
                }
                else {
                    glColor3f(1,1,1);
                    glVertex2i(x,y);
                }

            }
        }
        glEnd();
    }
    else if(DISPLAY_MODE == DTMAP) { // Display the distance transformed map
        glBegin(GL_POINTS);
        for(int y=0; y < DT.h; y++) {
            for(int x=0; x < DT.w; x++) {
                if(DT.DTmap[y*DT.w+x] == OBSTACLE) {
                    glColor3f(0,0,0);
                    glVertex2i(x,y);
                }
                else {
                    float v = DT.DTmap[y*DT.w+x] / DT.maxDTvalue; // normalise the value to be between [0..1]
                    glColor3f(0,v,0);
                    glVertex2i(x,y);
                }

            }
        }
        glEnd();
    }
    else if(DISPLAY_MODE == PFMAP) { // Display the potential field map
        glBegin(GL_POINTS);
        for(int y=0; y < DT.h; y++) {
            for(int x=0; x < DT.w; x++) {
                float v = DT.PFmap[y*DT.w+x] / DT.maxPFvalue; // normalise the value to be between [0..1]
                glColor3f(0,v,0);
                glVertex2i(x,y);
            }
        }
        glEnd();
    }

    // Draw the starting point
    glPointSize(8.0);

    glColor3f(1,0,0);
    glBegin(GL_POINTS);
    glVertex2i(DT.start.x, DT.start.y);
    glEnd();

    // Draw the goal point
    glColor3f(0,1,0);
    glBegin(GL_POINTS);
    glVertex2i(DT.goal.x, DT.goal.y);
    glEnd();

    // Draw the path
    glColor3f(1,0,1);
    glLineWidth(2.0);
    glBegin(GL_LINE_STRIP);
    for(unsigned int i=0; i < DT.path.size(); i++) {
        glVertex2i(DT.path[i].x, DT.path[i].y);
    }
    glEnd();

    if(!normalPath) { // Displaying the linearise path
        glPointSize(5);
        glBegin(GL_POINTS);
        for(unsigned int i=0; i < DT.path.size(); i++) {
            glVertex2i(DT.path[i].x, DT.path[i].y);
        }
        glEnd();
    }

    glutSwapBuffers();
}

void reshape(int w, int h)
{
    // Reset the coordinate system before modifying
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // Set the viewport to be the entire window
    glViewport(0, 0, w, h);

    // Set to 2D rendering
    gluOrtho2D(0, w, 0, h);

    SCREEN_W = w;
    SCREEN_H = h;
}

void idle()
{
    draw();
}

// Copy and paste from http://www.lighthouse3d.com/opengl/glut/index.php?bmpfont
void renderBitmapString(float x, float y, float z, void *font, const char *string)
{
    const char *c;
    glRasterPos3f(x, y,z);

    for (c=string; *c != '\0'; c++) {
        glutBitmapCharacter(font, *c);
    }
}

void keyboard(unsigned char key, int x, int y)
{
    switch(key) {
        case ' ': // spacebar, basic distance transform
        DT.CalcDistanceTransform(false);
        DT.FindPath();
        normalPath = true;
        break;

        case 's': // do the safe path version of the distance transform
        DT.CalcDistanceTransform(true);
        DT.FindPath();
        normalPath = true;
        break;

        case 'l': // linearise the path
        DT.FindLinearisePath();
        normalPath = false;
        break;

        case '1':
        DISPLAY_MODE = NORMAL;
        break;

        case '2':
        DISPLAY_MODE = DTMAP;
        break;

        case '3':
        DISPLAY_MODE = PFMAP;
        break;

        case 27: // ESC
        exit(0);
        break;
    }
}


void mouse(int button, int state, int x, int y)
{
	if(state == GLUT_DOWN) {
        if (button == GLUT_LEFT_BUTTON) { // change starting point
			DT.start = XY(x,SCREEN_H-y); // mouse co-ordinate is upside down relative to the display co-ordinate
        }
		else if (button == GLUT_RIGHT_BUTTON) {
			DT.goal = XY(x,SCREEN_H-y); // change goal point
			DT.CalcDistanceTransform(); // update the distance transform
		}
	}
}



